// 23- ( STL ,  )
// (c) Proteus     lawnmower-man@mail.ru

#include<stdio.h>
#include<stdlib.h>
#include<assert.h>

#define NODE_SIZE 3

inline int min(int a,int b) {return a<b?a:b;}

//!  
struct node_st{
	node_st(int);								
	int value;									//!<  
	node_st *next;								//!<  ,    
	node_st *child;								//!<  
	void Clear();								//!<  
	void Dump(int depth);						//!<  
	bool Add(node_st*);
	int GetCount();
	int GetMin();
};


//!  (    )
class Vector{
	node_st **array; int size;
	void Resize(int size);
public:
	Vector();
	~Vector();
	int GetSize();
	node_st *&operator[] (int index);
	bool isEmpty();
}; 

//!  
class _23heap{
	Vector vector;								//!<  
	void Insert(int h,node_st*);				//!<     
public:
	int  Pop();									//!<   
	void Insert(int v);							//!<    
	bool isEmpty();								//!<   
	void Dump();								//!<  
};



Vector::Vector()
{
	array=0; size=0;
}

Vector::~Vector()
{
	while(size--) if(array[size]) array[size]->Clear();
	free(array);
}

void Vector::Resize(int _size)
{
	int n=size;
	while(_size>=size) size+=1+size/4;
	if(n!=size) array=(node_st**)realloc(array,size*sizeof(node_st*));
	while(n<size) array[n++]=0;
}

node_st* &Vector::operator [](int index)
{
	Resize(index);
	return array[index];
}

bool Vector::isEmpty()
{
	for(int n=0; n<size; n++) if(array[n]) return false;
	return true;
}

int Vector::GetSize()
{
	return size;
}


//! \param v  
node_st::node_st(int v)
{
	value=v;
	next=child=0;
}


void node_st::Clear()
{
	if(next) next->Clear();
	if(child) child->Clear();
	delete this;
}

//! \param depth  
void node_st::Dump(int depth)
{
	int n;
	for(n=0; n<depth; n++) putchar(' ');
	printf("[%d]",value);
	if(child || !next) putchar('\n');
	if(child) child->Dump(depth+2);
	if(next)  next->Dump(child?depth:0);
}

int node_st::GetCount()
{
	int count=0;
	node_st *temp=this;
	while(temp) count++,temp=temp->next;
	assert(count>=2 && count<=NODE_SIZE);
	return count;
}

int node_st::GetMin()
{
	int min=value;
	node_st *temp=next;
	while(temp) {
		if(min>temp->value) min=temp->value;
		temp=temp->next;
	};
	return min;
}


//! \param list   
bool node_st::Add(node_st *list)
{
	assert(!next);
	next=list;
	if(GetCount()<NODE_SIZE) return false;
	int min=GetMin();
	node_st *node=new node_st(value);
	node->next=next; 
	node->child=child;
	next=0;
	child=node;
	value=min;
	return true;
}


//! \param v  
void _23heap::Insert(int v)
{
	Insert(0,new node_st(v));
}

//! \param item  
void _23heap::Insert(int h,node_st *item)
{
	item->next=0;
	node_st *add;
	bool up=true;
	for(;;) {
		add=vector[h]; if(!add) break;
		up=item->Add(add); if(!up) break;
		vector[h++]=0;
	}
	vector[h]=item;
}


bool _23heap::isEmpty()
{
	return vector.isEmpty();
}


int _23heap::Pop()
{
	bool unlock;	//     
	int n; int min,t;
	int index=-1;
	node_st *child,*next,*node;
	//  
	for(n=0; n<vector.GetSize(); n++) if(vector[n]) {
		t=vector[n]->GetMin();
		if(index<0 || t<min) min=t,index=n;
	}
	child=vector[index]; vector[index]=0;
	//     
	while(child) {
		node=child; child=0; unlock=true;
		while(node) {
			next=node->next; 
			if(unlock && node->value==min) {	//         
				unlock=false;
				child=node->child;
				delete node;
			} else {							//      ,   
				Insert(index,node);
			}
			node=next;
		}
		index--;
	}
	return min;
}


void _23heap::Dump()
{
	printf("[dump]\n");
	for(int n=0; n<vector.GetSize(); n++) if(vector[n]) {
		printf(" -root:%d-\n",n);
		vector[n]->Dump(0);
	}
}



#define SIZE 40


void main()
{
	_23heap heap;
	for(int n=0; n<SIZE; n++) heap.Insert(rand()%30);
	heap.Dump();
	while(!heap.isEmpty()) {
		printf("(%d)",heap.Pop()); 
	}
	putchar('\n');
	printf("[press enter]\n");
	getchar();
}
